﻿using FarPoint.Win.Spread;
using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace SpreadCustomActionsSample
{
  public partial class HierarchyNavigate : Form
  {
    public HierarchyNavigate()
    {
      InitializeComponent();
    }

    private void HierarchyNavigate_Load(object sender, EventArgs e)
    {
      fpSpread1.GetActionMap().Put("HierarchyMoveToPreviousRow", new HierarchyMoveToPreviousRowAction());
      fpSpread1.GetActionMap().Put("HierarchyMoveToNextRow", new HierarchyMoveToNextRowAction());
      InputMap inputMap = fpSpread1.GetInputMap(InputMapMode.WhenFocused);
      inputMap.Put(new Keystroke(Keys.Up, Keys.None), "HierarchyMoveToPreviousRow");
      inputMap.Put(new Keystroke(Keys.Down, Keys.None), "HierarchyMoveToNextRow");
      fpSpread1.DataSource = CreateDataSource();
      ExpandChildrenRecursive(fpSpread1_Sheet1);

      SpreadView root = fpSpread1.GetRootWorkbook();
      foreach (SpreadView child in root.GetChildWorkbooks())
        SetInputMapRecursive(inputMap, root, child);
    }

    private void SetInputMapRecursive(InputMap inputMap, SpreadView root, SpreadView spreadView)
    {
      spreadView.SetInputMap(InputMapMode.WhenFocused, inputMap);
      spreadView.GetActionMap().Put("HierarchyMoveToPreviousRow", new HierarchyMoveToPreviousRowAction());
      spreadView.GetActionMap().Put("HierarchyMoveToNextRow", new HierarchyMoveToNextRowAction());
      System.Diagnostics.Debug.WriteLine("Setting input map to view");
      inputMap.Put(new Keystroke(Keys.Up, Keys.None), "HierarchyMoveToPreviousRow");
      inputMap.Put(new Keystroke(Keys.Down, Keys.None), "HierarchyMoveToNextRow");
      ArrayList children = spreadView.GetChildWorkbooks();
      foreach (SpreadView child in children)
        SetInputMapRecursive(inputMap, root, child);
    }
    private void ExpandChildrenRecursive(SheetView sv)
    {
      sv.ExpandRow(-1, true);
      foreach (SheetView child in sv.GetChildSheets())
        ExpandChildrenRecursive(child);
    }
    private DataSet CreateDataSource()
    {
      DataSet ds = new DataSet();
      DataTable dtOffices = ds.Tables.Add("Offices");
      DataTable dtPeople = ds.Tables.Add("People");
      DataTable dtCars = ds.Tables.Add("Cars");
      DataTable dtPets = ds.Tables.Add("Pets");

      dtOffices.Columns.AddRange(new DataColumn[] { new DataColumn("Location"), new DataColumn("OfficeID", typeof(int)) });
      dtOffices.Rows.Add(new object[] { "Raleigh, NC", 0 });
      dtOffices.Rows.Add(new object[] { "Pittsburgh, PA", 1 });
      dtOffices.Rows.Add(new object[] { "Kirkland, WA", 2 });

      dtPeople.Columns.AddRange(new DataColumn[] { new DataColumn("FirstName"), new DataColumn("LastName"), new DataColumn("PersonID", typeof(int)), new DataColumn("OfficeID", typeof(int)) });
      dtPeople.Rows.Add(new object[] { "Sean", "Fieldes", 0, 0 });
      dtPeople.Rows.Add(new object[] { "Dan", "Smith", 1, 0 });
      dtPeople.Rows.Add(new object[] { "Bo", "Jones", 2, 0 });
      dtPeople.Rows.Add(new object[] { "Kevin", "Hill", 3, 0 });
      dtPeople.Rows.Add(new object[] { "Gina", "West", 4, 0 });
      dtPeople.Rows.Add(new object[] { "John", "Young", 5, 1 });
      dtPeople.Rows.Add(new object[] { "Andrew", "Perkins", 6, 1 });
      dtPeople.Rows.Add(new object[] { "Chris", "Davis", 7, 1 });
      dtPeople.Rows.Add(new object[] { "Eric", "Johnson", 8, 1 });
      dtPeople.Rows.Add(new object[] { "Justin", "Thomas", 9, 1 });
      dtPeople.Rows.Add(new object[] { "Devinder", "Lee", 10, 2 });
      dtPeople.Rows.Add(new object[] { "Sanjeev", "Clark", 11, 2 });
      dtPeople.Rows.Add(new object[] { "Alex", "Walker", 12, 2 });

      dtCars.Columns.AddRange(new DataColumn[] { new DataColumn("Make"), new DataColumn("Model"), new DataColumn("Color"), new DataColumn("PersonID", typeof(int)) });
      dtCars.Rows.Add(new object[] { "Honda", "CR-V", "Silver", 0 });
      dtCars.Rows.Add(new object[] { "Saab", "97-X", "Black", 0 });
      dtCars.Rows.Add(new object[] { "Jeep", "Cherokee", "Black", 1 });
      dtCars.Rows.Add(new object[] { "Ford", "Mustang", "Black", 2 });
      dtCars.Rows.Add(new object[] { "Volkswagon", "Beetle", "Black", 3 });
      dtCars.Rows.Add(new object[] { "Honda", "CR-V", "Blue", 4 });
      dtCars.Rows.Add(new object[] { "Ferari", "Testarosa", "Red", 4 });
      dtCars.Rows.Add(new object[] { "Ford", "F150", "Red", 5 });
      dtCars.Rows.Add(new object[] { "Honda", "Accord", "Silver", 6 });
      dtCars.Rows.Add(new object[] { "Toyota", "Tundra", "Black", 7 });
      dtCars.Rows.Add(new object[] { "Honda", "Odyssey", "Silver", 8 });
      dtCars.Rows.Add(new object[] { "Jeep", "Wrangler", "Silver", 9 });
      dtCars.Rows.Add(new object[] { "Mazarati", "Quatroport", "Black", 10 });
      dtCars.Rows.Add(new object[] { "Lexus", "LX", "Black", 11 });
      dtCars.Rows.Add(new object[] { "Honda", "Accord", "Black", 12 });

      dtPets.Columns.AddRange(new DataColumn[] { new DataColumn("Name"), new DataColumn("Type"), new DataColumn("Color"), new DataColumn("Gender"), new DataColumn("PersonID", typeof(int)) });
      dtPets.Rows.Add(new object[] { "Putzel", "Cat", "Orange", "Male", 0 });
      dtPets.Rows.Add(new object[] { "Hickory", "Cat", "Black", "Male", 0 });
      dtPets.Rows.Add(new object[] { "Ginger", "Cat", "Orange", "Female", 0 });
      dtPets.Rows.Add(new object[] { "Spice", "Cat", "Orange", "Male", 0 });
      dtPets.Rows.Add(new object[] { "Cupcake", "Cat", "Gray", "Male", 0 });
      dtPets.Rows.Add(new object[] { "Kitkat", "Cat", "Gray", "Male", 0 });
      dtPets.Rows.Add(new object[] { "Sweetie", "Cat", "Black/White Tabby", "Female", 0 });
      dtPets.Rows.Add(new object[] { "Fido", "Dog", "Black", "Male", 1 });
      dtPets.Rows.Add(new object[] { "Herbie", "Hamster", "Brown", 2 });
      dtPets.Rows.Add(new object[] { "Ricky", "Rhinoseros", "Black", 3 });
      dtPets.Rows.Add(new object[] { "Freddy", "Penguin", "Black and White", 4 });
      dtPets.Rows.Add(new object[] { "Kermit", "Frog", "Green", 5 });
      dtPets.Rows.Add(new object[] { "Fozzy", "Bear", "Brown", 6 });
      dtPets.Rows.Add(new object[] { "Tim", "Turtle", "Green", 7 });
      dtPets.Rows.Add(new object[] { "Wanda", "Fish", "Rainbow", 8 });
      dtPets.Rows.Add(new object[] { "Bruce", "Dog", "Brown", 9 });

      ds.Relations.Add(new DataRelation("OfficePerson", dtOffices.Columns["OfficeID"], dtPeople.Columns["OfficeID"]));
      ds.Relations.Add(new DataRelation("PersonCar", dtPeople.Columns["PersonID"], dtCars.Columns["PersonID"]));
      ds.Relations.Add(new DataRelation("PersonPet", dtPeople.Columns["PersonID"], dtPets.Columns["PersonID"]));
      return ds;
    }

    private class HierarchyMoveToPreviousRowAction : FarPoint.Win.Spread.Action
    {
      private SpreadView FindLastChildRecursive(SpreadView spreadView)
      {
        ArrayList children = spreadView.GetChildWorkbooks();
        if (children == null || children.Count == 0)
          return spreadView;
        else
        {
          for (int i = children.Count - 1; i >= 0; i--)
          {
            SheetView sheet = ((SpreadView)children[i]).GetSheetView();
            if (sheet.RowCount > 0)
              return FindLastChildRecursive(((SpreadView)children[i]));
          }
        }
        return spreadView;
      }
      public override void PerformAction(object sender)
      {
        SpreadView spreadView = sender as SpreadView;
        if (spreadView != null)
        {
          SheetView sheetView = spreadView.GetSheetView();
          int activeColumn = sheetView.ActiveColumnIndex;
          int activeRow = sheetView.ActiveRowIndex;
          int parentRowIndex = sheetView.ParentRowIndex;
          SpreadView parentSpreadView = spreadView.Parent;
          if (parentSpreadView != null)
          { // spreadView is a child workbook
            ArrayList siblings = parentSpreadView.GetChildWorkbooks();
            ArrayList children = spreadView.GetChildWorkbooks();
            int childViewIndex = siblings.IndexOf(spreadView);
            if (activeRow == 0 && childViewIndex > 0 && ((SpreadView)siblings[childViewIndex - 1]).GetSheetView().ParentRowIndex == parentRowIndex)
            { // there is a sibling sheet above this one, go to its last row
              SpreadView targetSpreadView = (SpreadView)siblings[childViewIndex - 1];
              SheetView targetSheetView = targetSpreadView.GetSheetView();
              spreadView.GetRootWorkbook().SetActiveWorkbook(targetSpreadView);
              targetSheetView.SetActiveCell(targetSheetView.RowCount - 1, activeColumn);
              targetSpreadView.ShowCell(targetSpreadView.GetActiveRowViewportIndex(), targetSpreadView.GetActiveColumnViewportIndex(), targetSheetView.ActiveRowIndex, targetSheetView.ActiveColumnIndex, VerticalPosition.Nearest, HorizontalPosition.Nearest);
              return;
            }
            else
            {
              SheetView targetSheetView = null;
              if (activeRow > 0 && sheetView.ChildRelationCount > 0)
              { // there are child sheets for the previous row, go to the last child's last row (unless they're all empty)
                for (int n = sheetView.ChildRelationCount - 1; n >= 0; n--)
                {
                  targetSheetView = sheetView.GetChildView(activeRow - 1, n);
                  if (targetSheetView.RowCount > 0)
                    break;
                }
                if (targetSheetView.RowCount > 0)
                {
                  SpreadView targetSpreadView = targetSheetView.ContainingViews[0];
                  spreadView.GetRootWorkbook().SetActiveWorkbook(targetSpreadView);
                  targetSheetView.SetActiveCell(targetSheetView.RowCount - 1, targetSheetView.ActiveColumnIndex);
                  targetSpreadView.ShowCell(targetSpreadView.GetActiveRowViewportIndex(), targetSpreadView.GetActiveColumnViewportIndex(), targetSheetView.ActiveRowIndex, targetSheetView.ActiveColumnIndex, VerticalPosition.Nearest, HorizontalPosition.Nearest);
                  return;
                }
              }
              if (activeRow == 0)
              { // there is no sibling sheet above this one, and no child sheets for the previous row, go to the parent row
                spreadView.GetRootWorkbook().SetActiveWorkbook(parentSpreadView);
                targetSheetView = parentSpreadView.GetSheetView();
                targetSheetView.SetActiveCell(sheetView.ParentRowIndex, targetSheetView.ActiveColumnIndex);
                parentSpreadView.ShowCell(parentSpreadView.GetActiveRowViewportIndex(), parentSpreadView.GetActiveColumnViewportIndex(), targetSheetView.ActiveRowIndex, targetSheetView.ActiveColumnIndex, VerticalPosition.Nearest, HorizontalPosition.Nearest);
                return;
              }
            }
          }
          else
          { // spreadView is top-level view (root workbook)
            ArrayList children = spreadView.GetChildWorkbooks();
            foreach (SpreadView child in children)
              if (child.GetSheetView().ParentRowIndex == activeRow - 1)
              {
                SpreadView targetSpreadView = FindLastChildRecursive(child);
                SheetView targetSheetView = targetSpreadView.GetSheetView();
                spreadView.GetRootWorkbook().SetActiveWorkbook(targetSpreadView);
                targetSheetView.SetActiveCell(targetSheetView.RowCount - 1, targetSheetView.ActiveColumnIndex);
                targetSpreadView.ShowCell(targetSpreadView.GetActiveRowViewportIndex(), targetSpreadView.GetActiveColumnViewportIndex(), targetSheetView.ActiveRowIndex, targetSheetView.ActiveColumnIndex, VerticalPosition.Nearest, HorizontalPosition.Nearest);
                return;
              }
          }
          // do default action
          spreadView.GetRootWorkbook().GetActionMap().Get("MoveToPreviousRow").PerformAction(sender);
        }
      }
    }

    private class HierarchyMoveToNextRowAction : FarPoint.Win.Spread.Action
    {
      private SpreadView FindFirstChildRecursive(SpreadView spreadView)
      {
        ArrayList children = spreadView.GetChildWorkbooks();
        if (children == null || children.Count == 0)
          return spreadView;
        else
        {
          for (int i = 0; i < children.Count - 1; i++)
          {
            SheetView sheet = ((SpreadView)children[i]).GetSheetView();
            if (sheet.RowCount > 0)
              return FindFirstChildRecursive(((SpreadView)children[i]));
          }
        }
        return spreadView;
      }
      public override void PerformAction(object sender)
      {
        SpreadView spreadView = sender as SpreadView;
        if (spreadView != null)
        {
          SheetView sheetView = spreadView.GetSheetView();
          int activeColumn = sheetView.ActiveColumnIndex;
          int activeRow = sheetView.ActiveRowIndex;
          int parentRowIndex = sheetView.ParentRowIndex;
          SpreadView parentSpreadView = spreadView.Parent;
          if (parentSpreadView != null)
          { // spreadView is a child workbook
            SpreadView targetSpreadView = null;
            SheetView targetSheetView = null;
            ArrayList siblings = parentSpreadView.GetChildWorkbooks();
            ArrayList children = spreadView.GetChildWorkbooks();
            int childViewIndex = siblings.IndexOf(spreadView);
            if (activeRow == sheetView.RowCount - 1 && childViewIndex < siblings.Count - 1 && ((SpreadView)siblings[childViewIndex + 1]).GetSheetView().ParentRowIndex == parentRowIndex)
            { // there is a sibling sheet below this one, go to its first row (unless they're all empty)
              for( int n = childViewIndex + 1; n < siblings.Count && ((SpreadView)siblings[n]).GetSheetView().ParentRowIndex == parentRowIndex; n++)
              {
                targetSpreadView = (SpreadView)siblings[n];
                targetSheetView = targetSpreadView.GetSheetView();
                if (targetSheetView.RowCount > 0)
                  break;
              }
              if (targetSheetView.RowCount > 0)
              {
                spreadView.GetRootWorkbook().SetActiveWorkbook(targetSpreadView);
                targetSheetView.SetActiveCell(0, activeColumn);
                targetSpreadView.ShowCell(targetSpreadView.GetActiveRowViewportIndex(), targetSpreadView.GetActiveColumnViewportIndex(), targetSheetView.ActiveRowIndex, targetSheetView.ActiveColumnIndex, VerticalPosition.Nearest, HorizontalPosition.Nearest);
                targetSheetView.Cells[targetSheetView.ActiveRowIndex, targetSheetView.ActiveColumnIndex].Invalidate();
                return;
              }
            }
            if (sheetView.ChildRelationCount > 0)
            { // there are child sheets for the next row, go to the first child's first row (unless they're all empty)
              for (int n = 0; n < sheetView.ChildRelationCount - 1; n++)
              {
                targetSheetView = sheetView.GetChildView(activeRow, n);
                if (targetSheetView.RowCount > 0)
                  break;
              }
              if (targetSheetView.RowCount > 0)
              {
                targetSpreadView = targetSheetView.ContainingViews[0];
                spreadView.GetRootWorkbook().SetActiveWorkbook(targetSpreadView);
                targetSheetView.SetActiveCell(0, targetSheetView.ActiveColumnIndex);
                targetSpreadView.ShowCell(targetSpreadView.GetActiveRowViewportIndex(), targetSpreadView.GetActiveColumnViewportIndex(), targetSheetView.ActiveRowIndex, targetSheetView.ActiveColumnIndex, VerticalPosition.Nearest, HorizontalPosition.Nearest);
                return;
              }
            }
            if (activeRow == sheetView.RowCount - 1)
            { // there is no sibling sheet below this one, and no child sheets for the next row, go to the next parent row
              targetSheetView = parentSpreadView.GetSheetView();
              while( parentSpreadView != null && parentRowIndex + 1 >= targetSheetView.RowCount )
              { // no more rows at this level, go up one more level
                parentRowIndex = targetSheetView.ParentRowIndex;
                parentSpreadView = parentSpreadView.Parent;
                targetSheetView = parentSpreadView == null ? null : parentSpreadView.GetSheetView();
              }
              if (parentSpreadView != null)
              {
                spreadView.GetRootWorkbook().SetActiveWorkbook(parentSpreadView);
                targetSheetView.SetActiveCell(parentRowIndex + 1, targetSheetView.ActiveColumnIndex);
                parentSpreadView.ShowCell(parentSpreadView.GetActiveRowViewportIndex(), parentSpreadView.GetActiveColumnViewportIndex(), targetSheetView.ActiveRowIndex, targetSheetView.ActiveColumnIndex, VerticalPosition.Nearest, HorizontalPosition.Nearest);
                return;
              }
            }
          }
          else
          { // spreadView is top-level view (root workbook)
            ArrayList children = spreadView.GetChildWorkbooks();
            foreach (SpreadView child in children)
              if (child.GetSheetView().ParentRowIndex == activeRow)
              {
                SheetView targetSheetView = child.GetSheetView();
                spreadView.GetRootWorkbook().SetActiveWorkbook(child);
                targetSheetView.SetActiveCell(0, targetSheetView.ActiveColumnIndex);
                child.ShowCell(child.GetActiveRowViewportIndex(), child.GetActiveColumnViewportIndex(), targetSheetView.ActiveRowIndex, targetSheetView.ActiveColumnIndex, VerticalPosition.Nearest, HorizontalPosition.Nearest);
                return;
              }
          }
          // do default action
          spreadView.GetRootWorkbook().GetActionMap().Get("MoveToNextRow").PerformAction(sender);
        }
      }
    }
  }


}
